﻿/*	FT232-LCD 制御ライブラリ
	『昼夜逆転』工作室	http://jsdiy.web.fc2.com

namespace FT232LcdLib
class LcdCtrl

【コンストラクタ】

LcdCtrl(FTDI ftdev, BitBangSignal lowestDataBit, BitBangSignal bitRS, BitBangSignal bitE)

FTDI ftdev					RS232デバイスのオブジェクト
BitBangSignal lowestDataBit	データ線の最下位ビット（例：D3-D0使用ならD0、D7-D4使用ならD4）
BitBangSignal bitRS			LCDモジュールのRegister Select線の割り当て
BitBangSignal bitE			LCDモジュールのEnable Signal線の割り当て

【プロパティ】

Is8x2to16x1
bool {get|set}
	見かけ上16文字x1行だが内部は8文字x2行というLCDモジュールに対して（例:S-10551D）、
	見かけ通り16文字x1行と扱えるようにするフラグ（静止表示の場合）。	
	※もともと見かけ上が8文字x2行のLCDモジュールに対しては、自動改行する振る舞いになる。

【メソッド】

FT232のビット出力をリセットする
void BitReset()
	データ線4ビット＋RS線＋E線 の6線に0(Low)を出力する。

LCDコマンド実行
void Cmd(int data)
int data	LCDコマンド。LcdCmd,LcdSubCmd定義の値。またはそれらと同等の整数。

LCD初期化
void Init()
	コンストラクタでLcdCtrlオブジェクトを生成後、最初に実行すること。一度実行すれば十分。

LCD画面消去
void Cls()

表示位置指定
void SetPos(int row, int column)
int row		行 0,1
int column	桁 0,1,2,…,15
	引数の具体値は2行16桁LCDモジュールの例。8桁のLCDモジュールであればcolumn=0～7が有効。

1文字出力する
void PutChar(char c)
void PutChar(byte c)
	引数はLCDモジュールで定義された文字コード。
	.NET Frameworkのchar型はUnicodeなので1バイトではないことに注意（'a'=0x0061 'あ'=0x3042）。
	※もとよりマルチバイト文字は受け付けていない。

文字列を出力する
void PutStr(string s)
void PutStr(string s, int row, int column)

数値を出力する
void PutInt(int n)
void PutInt(int n, int row, int column)

外字を定義する
void DefChar(int id, byte[] dots)
int    id		0～7（キャラクタコード）
byte[] dots		5x8のドットパターンを格納した配列


【更新履歴】
2011/09/24	ver 1.0.0	初版。
*/


using System;
using System.Text;			//Encoding
using System.Threading;		//Sleep
using FTD2XX_NET;

namespace FT232LcdLib
{
	//enumは演算に使用するとき明示的にint型にキャストしなければならない
	//それが面倒なのでstructで定義している
	
	//Instruction
	struct LcdCmd
	{
		public const int
			CLEAR_DISPLAY	= 0x01,	//0b00000001
			RETURN_HOME		= 0x02,	//0b00000010
			ENTRY_MODE		= 0x04,	//0b00000100
			DISPLAY_MODE	= 0x08,	//0b00001000
			CURSOR_MODE		= 0x10,	//0b00010000
			FUNCTION_SET	= 0x20,	//0b00100000
			SET_CGRAM_ADRS	= 0x40,	//0b01000000
			SET_DDRAM_ADRS	= 0x80;	//0b10000000
	}
	
	//Instructionのパラメータ
	//(LCD_CMD.AAA | LCD_SUBCMD.XXX1 | LCD_SUBCMD.XXX2 | …)の形で使う
	struct LcdSubCmd
	{
		public const int
			//LCD_CMD.ENTRY_MODE
			INCREMENT		= 0x02,	//0b00000010	//入力位置を進める（無指定だとDecrement）
			WITH_SHIFT		= 0x01,	//0b00000001	//表示シフト動作を伴う（LCD_CMD.CURSOR_MODEも設定する）
			//LCD_CMD.DISPLAY_MODE
			SET_DISPLAY		= 0x04,	//0b00000100	//画面表示する（無指定だと表示内容を保ったまま画面を消す）
			SET_CURSOR		= 0x02,	//0b00000010	//カーソルを表示する（アンダーバーを表示する）
			CURSOR_BLINK	= 0x01,	//0b00000001	//カーソルを点滅させる（文字に被さって黒マスが点滅する）
			//LCD_CMD.CURSOR_MODE
			SHIFT_ON		= 0x08,	//0b00001000	//DDRAMの変更無しに表示をシフトさせる（無指定だとカーソルを移動させる）
			RIGHT			= 0x04,	//0b00000100	//右へシフトさせる（無指定だと左へ）※上記、カーソル移動モード時は意味を持たない
			//LCD_CMD.FUNCTION_SET
			DL8BIT			= 0x10,	//0b00010000	//Data Length 8-bit（無指定だと4-bitモード）
			LINE2			= 0x08,	//0b00001000	//表示で2行使用する（無指定だと1行のみ使用可能になる）
			FONT5X10		= 0x04;	//0b00000100	//Font Type 5x10（無指定だと5x8）※SC1602/SD1602は物理的に非対応
	}
	
	//FT232 ビットバンモードの信号名（各ビットの名前）
	enum BitBangSignal
	{
		D0, D1, D2, D3, D4, D5, D6, D7
	};
	
	/*	秋月電子FT232RLモジュールのピン配置
	   +----------+
	D0 |TXD    GND|
	D4 |DTR#   CB0|
	D2 |RTS#   CB1|
	   |VIO    VCC|
	D1 |RXD    RST|
	D7 |RI#    3V3|
	   |GND    CB3|
	D5 |DSR#   PU1|
	D6 |DCD#   PU2|
	D3 |CTS#   VCC|
	   |CB4    USB|(5V)
	   |CB2    SLD|(GND)
	   +----↓----+
	   USBコネクタ側
	*/
	
	
	class LcdCtrl
	{
		//LCDを制御するビットの割り当て（4bitモードが前提）
		//※LCDのRead/Write制御線は使用せず（ハード的にwrite固定とする）
		private readonly int
			LCD_RS,					//Register Select	H:data	L:command
			LCD_E,					//Enable Signal		H->L:strobe
			LCD_DATABITS,			//データ線4bit幅
			LCD_DATABITS_SHIFT;		//データ線の最下位ビット（例：D3-D0使用ならD0、D7-D4使用ならD4）
		
		private readonly FTDI ftdev;
		private uint writtenLength = 0;
		private byte[] wbuf = new byte[1];
		private byte[] cbuf = new byte[5];
		private static readonly Encoding sjis = Encoding.GetEncoding("shift_jis");	//LCDモジュール側の文字コード
		
		private int columnIndex = 0;
		private bool is8x2to16x1 = false;
		
		//プロパティ
		public bool Is8x2to16x1
		{
			get { return is8x2to16x1; }
			set { is8x2to16x1 = value; SetPos(0, 0); }
		}
		
		
		//コンストラクタ
		public LcdCtrl(FTDI ftdev, BitBangSignal lowestDataBit, BitBangSignal bitRS, BitBangSignal bitE)
		{
			this.ftdev = ftdev;
			LCD_RS = (1 << (int)bitRS);
			LCD_E  = (1 << (int)bitE);
			LCD_DATABITS_SHIFT = (int)lowestDataBit;
			LCD_DATABITS = (0x0F << LCD_DATABITS_SHIFT);
		}
		
		private void DelayMs(int ms)
		{
			Thread.Sleep(ms);
		}
		
		private void DelayUs(int us)
		{
			Thread.Sleep(1);
		}
		
		//1byteを出力する
		private bool Write(int byteData)
		{
			wbuf[0] = (byte)byteData;
			return (ftdev.Write(wbuf, 1, ref writtenLength) == FTDI.FT_STATUS.FT_OK);
		}
		
		//byte配列を出力する
		private bool Write(byte[] datas)
		{
			return (ftdev.Write(datas, datas.Length, ref writtenLength) == FTDI.FT_STATUS.FT_OK);
		}
		
		//FT232のビット出力をリセットする
		public void BitReset()
		{
			//LCD制御に使用していないピンの状態に影響を与えないよう配慮する
			byte pinState = FTDI.FT_BIT_MODES.FT_BIT_MODE_ASYNC_BITBANG;
			ftdev.GetPinStates(ref pinState);
			Write(pinState & ~(LCD_E | LCD_RS | LCD_DATABITS));
		}
		
		//LCDコマンド実行
		public void Cmd(int data)
		{
			Cmd4bit((data >> 4) & 0x0F);	//上位4bit
			Cmd4bit(data & 0x0F);			//下位4bit
		}
		
		private void Cmd4bit(int data)
		{
			//dataの内容は4bit幅であることが前提
			int bits = LCD_E | (data << LCD_DATABITS_SHIFT);
			Write(bits);
			DelayMs(1);
			
			Write(bits ^ LCD_E);
			DelayMs(1);
		}
		
		//LCD初期化
		public void Init()
		{
			DelayMs(15);
			Cmd4bit(0x03);	//0b0011
			DelayMs(5);
			Cmd4bit(0x03);	//0b0011
			DelayUs(100);
			Cmd4bit(0x03);	//0b0011
			
			//4-bitモードへ移行
			Cmd4bit(0x02);	//0b0010
			
			//----------bit:76543210
			Cmd(0x28);	//0b00101000	//Function Set	bit:5(1), DL=4-bit(0), N=2lines(1), F=5x8dots(0), *(0), *(0)
			Cmd(0x08);	//0b00001000	//Display Set	bit:3(1), D=Set Display:NO(0), C=Set Cursor:NO(0), B=Cursor Blink:NO(0)
			Cmd(0x01);	//0b00000001	//Clear Display	bit:0(1)
			Cmd(0x06);	//0b00000110	//Entry Mode	bit:2(1), I/D=Increment(1), S=With Display Shift:NO(0)
			Cmd(0x0C);	//0b00001100	//Display Set	bit:3(1), D=Set Display:YES(1), C=Set Cursor:NO(0), B=Cursor Blink:NO(0)
		}
		
		//LCD画面消去
		public void Cls()
		{
			Cmd(LcdCmd.CLEAR_DISPLAY);
		}
		
		//表示位置指定		※↓具体値は2行16桁LCDモジュールの例
		//引数	row:	行 0,1
		//		column:	桁 0,1,2,…,15
		public void SetPos(int row, int column)
		{
			if (row < 0 || column < 0) return;
			Cmd(LcdCmd.SET_DDRAM_ADRS | (row * 0x40 + column));
			columnIndex = column;
		}
		
		//1文字出力する
		public void PutChar(char c)
		{
			//文字の内容を無視してUnicodeから1byteのデータに変換する
			//※ファイル先頭コメント内、PutChar()の説明を参照のこと。
			//※Unicode'0x00??'の文字はbyte型にキャストしても影響を受けないのでこのような実装にしている。
			PutChar((byte)c);
		}
		
		//1文字出力する
		public void PutChar(byte c)
		{
			if (is8x2to16x1)
			{
				//columnIndex  :0-7/8-15/16-23/24-31/...
				//SetPos-row   :0  /1   /0    /1    /...
				//SetPos-column:0-7/0-7 /0-7  /0-7  /...
				if (columnIndex % 8 == 0) SetPos((columnIndex / 8) % 2, columnIndex % 8);
				columnIndex++;
			}
			
			int bits = LCD_RS | LCD_E | (((c >> 4) & 0x0F) << LCD_DATABITS_SHIFT);	//上位4bit
			cbuf[0] = (byte)bits;
			cbuf[1] = (byte)(bits ^ LCD_E);
			bits = LCD_RS | LCD_E | ((c & 0x0F) << LCD_DATABITS_SHIFT);		//下位4bit
			cbuf[2] = (byte)bits;
			bits ^= LCD_E;
			cbuf[3] = (byte)bits;
			cbuf[4] = (byte)(bits ^ LCD_RS);
			foreach (byte cc in cbuf)
				Write(cc);
		}
		
		//文字列を出力する
		public void PutStr(string s)
		{
			if (s == null) return;
			byte[] bytes = sjis.GetBytes(s);
			foreach (byte cc in bytes)
				PutChar(cc);
			bytes = null;
			GC.Collect();
		}
		
		//文字列を出力する
		public void PutStr(string s, int row, int column)
		{
			SetPos(row, column);
			PutStr(s);
		}
		
		//数値を出力する
		public void PutInt(int n)
		{
			PutStr(n.ToString());
		}
		
		//数値を出力する
		public void PutInt(int n, int row, int column)
		{
			SetPos(row, column);
			PutStr(n.ToString());
		}
		
		//外字を定義する
		//引数	id:		0～7（キャラクタコード）
		//		dots:	5x8のドットパターンを格納した配列
		public void DefChar(int id, byte[] dots)
		{
			if (id < 0 || 7 < id) return;
			if (dots == null || dots.Length < 8) return;
			for (int i = 0; i < 8; i++)
			{
				Cmd(LcdCmd.SET_CGRAM_ADRS | (id << 3) | i);
				PutChar(dots[i]);
			}
		}
		
	}	//<- end of class

}	//<- end of namespace
